# flake8: noqa

from hammurabi.config import config
from hammurabi.law import Law
from hammurabi.pillar import Pillar
from hammurabi.preconditions.attributes import (
    HasMode,
    HasNoMode,
    IsNotOwnedBy,
    IsOwnedBy,
)
from hammurabi.preconditions.base import Precondition
from hammurabi.preconditions.directories import IsDirectoryExists, IsDirectoryNotExists
from hammurabi.preconditions.files import IsFileExists, IsFileNotExists
from hammurabi.preconditions.text import IsLineExists, IsLineNotExists
from hammurabi.reporters.base import Reporter
from hammurabi.reporters.json import JSONReporter
from hammurabi.rules.attributes import ModeChanged, OwnerChanged
from hammurabi.rules.base import Rule
from hammurabi.rules.directories import (
    DirectoryEmptied,
    DirectoryExists,
    DirectoryNotExists,
)
from hammurabi.rules.files import (
    FileEmptied,
    FileExists,
    FileNotExists,
    FilesExist,
    FilesNotExist,
)
from hammurabi.rules.ini import (
    OptionRenamed,
    OptionsExist,
    OptionsNotExist,
    SectionExists,
    SectionNotExists,
    SectionRenamed,
)
from hammurabi.rules.json import (
    JSONKeyExists,
    JSONKeyNotExists,
    JSONKeyRenamed,
    JSONValueExists,
    JSONValueNotExists,
)
from hammurabi.rules.operations import Copied, Moved, Renamed
from hammurabi.rules.templates import TemplateRendered
from hammurabi.rules.text import LineExists, LineNotExists, LineReplaced
from hammurabi.rules.yaml import (
    YAMLKeyExists,
    YAMLKeyNotExists,
    YAMLKeyRenamed,
    YAMLValueExists,
    YAMLValueNotExists,
)

__version__ = "0.7.1"



# pylint: disable=too-few-public-methods
import logging

from hammurabi.reporters.base import Reporter


class JSONReporter(Reporter):
    """
    Generate reports in JSON format and write into file. JSONReporter
    is the default reporter of the pillar. The example below shows the
    way how to replace a reporter which could base on the JSONReporter.

    The report will be written into the configured report file. The report
    file's name set by ``report_name`` config parameter.

    Example usage:

    .. code-block:: python

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, OwnerChanged
        >>> from my_company import MyJSONReporter
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         OwnerChanged(
        >>>             name="Change ownership of nginx config",
        >>>             path=Path("./nginx.conf"),
        >>>             new_value="www:web_admin"
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> # override pillar's default JSONReporter reporter
        >>> pillar = Pillar(reporter_class=MyJSONReporter)
    """

    def report(self) -> None:
        """
        Do the actual reporting based on the report assembled in JSON
        format. The report will be written into the configured report file.
        """

        logging.info('Writing report to "%s"', str(self.report_path))
        self.report_path.write_text(self._get_report().json())



"""
Extend :class:`hammurabi.rules.base.Rule` to handle parsed content manipulations dictionaries.
Standalone these rules are not useful, but they are very handy when files should be manipulated
like YAML or JSON which will be parsed as dict.

These rules are intentionally not exported directly through hammurabi as it is done for YAML or
JSON rules. The reason, as it is mentioned above, these rules are not standalone rules. Also, it
is intentional that these rules are not represented in the documentation's `Rules section`_.

.. _`Rules section`: https://hammurabi.readthedocs.io/en/latest/rules.html
"""

from abc import ABC, abstractmethod
from copy import deepcopy
import logging
from pathlib import Path
from typing import Any, Callable, Dict, Hashable, List, Optional, Union

from hammurabi.rules.common import SinglePathRule
from hammurabi.rules.mixins import SelectorMixin


class SinglePathDictParsedRule(SinglePathRule, SelectorMixin):
    """
    Extend :class:`hammurabi.rules.base.Rule` to handle parsed content
    manipulations dictionaries. Standalone this rule is not useful, but
    it is very handy when files should be manipulated like YAML or
    JSON which will be parsed as dict. This rule ensures that the implementation
    will be the same for these rules, so the maintenance cost and effort
    is reduced.

    Although this rule is not that powerful on its own, we would not
    like to make it an abstract class like :class:`hammurabi.rules.base.Rule`
    because it can easily happen that at some point this rule will be
    a standalone rule.
    """

    def __init__(
        self,
        name: str,
        path: Optional[Path] = None,
        key: str = "",
        loader: Callable[[Any], Dict[str, Any]] = dict,
        **kwargs,
    ) -> None:
        self.selector = self.validate(key, required=True)
        self.split_key = self.selector.split(".")
        self.key_name: str = self.split_key[-1]
        self.loaded_data = Union[Dict[Hashable, Any], List[Any], None]
        self.loader = loader

        super().__init__(name, path, **kwargs)

    def _get_parent(self) -> Dict[str, Any]:
        """
        Get the parent of the given key by its selector.

        :return: Return the parent if there is any
        :rtype: Dict[str, Any]
        """

        # Get the parent for modifications. If there is no parent,
        # then the parent is the document root
        return self.get_by_selector(self.loaded_data, self.split_key[:-1])

    def _write_dump(self, data: Any, delete: bool = False) -> None:
        """
        This is a dummy class which should be overridden. This method
        does nothing.

        :param data: The modified data
        :type data: :class:``hammurabi.rules.mixins.Any`

        :param delete: Indicate if the key should be deleted
        :type delete: bool
        """

    def pre_task_hook(self) -> None:
        """
        Parse the file for later use.
        """

        logging.debug('Parsing "%s" file', self.param)
        self.loaded_data = self.loader(self.param.read_text())

    @abstractmethod
    def task(self) -> Path:
        """
        Abstract method representing how a :func:`hammurabi.rules.base.Rule.task`
        must be parameterized. Any difference in the parameters will result in
        pylint/mypy errors.

        For more details please check :func:`hammurabi.rules.base.Rule.task`.
        """


class DictKeyExists(SinglePathDictParsedRule, ABC):
    """
    Ensure that the given key exists. If needed, the rule will create a key with the
    given name, and optionally the specified value. In case the value is set, the value
    will be assigned to the key. If no value is set, the key will be created with an empty
    value.

    Example usage:

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar
        >>> from hammurabi.rules.dictionaries import DictKeyExists
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         DictKeyExists(
        >>>             name="Ensure service descriptor has stack",
        >>>             path=Path("./service.dictionary"),
        >>>             key="stack",
        >>>             value="my-awesome-stack",
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)

    .. warning::

        Compared to :mod:`hammurabi.rules.text.LineExists`, this rule is NOT able to add a
        key before or after a target.
    """

    def __init__(
        self,
        name: str,
        path: Optional[Path] = None,
        key: str = "",
        value: Union[None, list, dict, str, int, float] = None,
        **kwargs,
    ) -> None:
        self.value = value
        super().__init__(name, path, key, **kwargs)

    def task(self) -> Path:
        """
        Ensure that the given key exists in the parsed file. If needed, create the
        key with the given name, and optionally the specified value.

        :return: Return the input path as an output
        :rtype: Path
        """

        parent = self._get_parent()

        logging.debug(
            'Set default value "%s" for "%s" if no value set', self.key_name, self.value
        )
        inserted = parent.setdefault(self.key_name, self.value)

        # Only write the changes if we did any change
        if inserted == parent[self.key_name]:
            self._write_dump(inserted)

        return self.param


class DictKeyNotExists(SinglePathDictParsedRule, ABC):
    """
    Ensure that the given key not exists. If needed, the rule will remove a key with the
    given name, including its value.

    Example usage:

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar
        >>> from hammurabi.rules.dictionaries import DictKeyNotExists
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         DictKeyNotExists(
        >>>             name="Ensure outdated_key is removed",
        >>>             path=Path("./service.dictionary"),
        >>>             key="outdated_key",
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)
    """

    def task(self) -> Path:
        """
        Ensure that the given key does not exists in the parsed file.

        :return: Return the input path as an output
        :rtype: Path
        """

        parent = self._get_parent()

        if self.key_name in parent.keys():
            logging.debug('Removing key "%s"', self.key_name)
            parent.pop(self.key_name)
            self._write_dump(parent, delete=True)

        return self.param


class DictKeyRenamed(SinglePathDictParsedRule, ABC):
    """
    Ensure that the given key is renamed. In case the key can not be found,
    a ``LookupError`` exception will be raised to stop the execution. The
    execution must be stopped at this point, because if other rules depending
    on the rename they will fail otherwise.

    Example usage:

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar
        >>> from hammurabi.rules.dictionaries import DictKeyRenamed
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         DictKeyRenamed(
        >>>             name="Ensure service descriptor has dependencies",
        >>>             path=Path("./service.dictionary"),
        >>>             key="development.depends_on",
        >>>             value="dependencies",
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)
    """

    def __init__(
        self,
        name: str,
        path: Optional[Path] = None,
        key: str = "",
        new_name: str = "",
        **kwargs,
    ) -> None:
        self.new_name = self.validate(new_name, required=True)
        super().__init__(name, path, key, **kwargs)

    def task(self) -> Path:
        """
        Ensure that the given key is renamed. In case the key can not be found,
        a ``LookupError`` exception will be raised to stop the execution. The
        execution must be stopped at this point, because if other rules depending
        on the rename they will fail otherwise.

        :raises: ``LookupError`` raised if no key can be renamed or both the new and
                 old keys are in the config file
        :return: Return the input path as an output
        :rtype: Path
        """

        parent = self._get_parent()

        has_old_key = self.key_name in parent
        has_new_key = self.new_name in parent

        if has_old_key and has_new_key:
            raise LookupError(f'Both "{self.key_name}" and "{self.new_name}" set')

        if has_new_key:
            return self.param

        if not has_old_key:
            raise LookupError(f'No matching key for "{self.selector}"')

        logging.debug('Renaming key from "%s" to "%s"', self.key_name, self.new_name)
        parent[self.new_name] = deepcopy(parent[self.key_name])
        parent.pop(self.key_name)

        # Delete is True since we need to delete the old key
        self._write_dump(parent, delete=True)

        return self.param


class DictValueExists(SinglePathDictParsedRule, ABC):
    """
    Ensure that the given key has the expected value(s). In case the key cannot
    be found, a ``LookupError`` exception will be raised to stop the execution.

    This rule is special in the way that the value can be almost anything. For
    more information please read the warning below.

    Example usage:

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar
        >>> from hammurabi.rules.dictionaries import DictValueExists
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         DictValueExists(
        >>>             name="Ensure service descriptor has dependencies",
        >>>             path=Path("./service.dictionary"),
        >>>             key="development.dependencies",
        >>>             value=["service1", "service2", "service3"],
        >>>         ),
        >>>         # Or
        >>>         DictValueExists(
        >>>             name="Add infra alerting to existing alerting components",
        >>>             path=Path("./service.dictionary"),
        >>>             key="development.alerting",
        >>>             value={"infra": "#slack-channel-2"},
        >>>         ),
        >>>         # Or
        >>>         DictValueExists(
        >>>             name="Add support info",
        >>>             path=Path("./service.dictionary"),
        >>>             key="development.supported",
        >>>             value=True,
        >>>         ),
        >>>         # Or even
        >>>         DictValueExists(
        >>>             name="Make sure that no development branch is set",
        >>>             path=Path("./service.dictionary"),
        >>>             key="development.branch",
        >>>             value=None,
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)

    .. warning::

        Since the value can be anything from ``None`` to a list of lists, and
        rule piping passes the 1st argument (``path``) to the next rule the ``value``
        parameter can not be defined in ``__init__`` before the ``path``. Hence
        the ``value`` parameter must have a default value. The default value is
        set to ``None``, which translates to the following:

        Using the ``DictValueExists`` rule and not assigning value to ``value``
        parameter will set the matching ``key``'s value to `None`` by default in
        the document.
    """

    def __init__(
        self,
        name: str,
        path: Optional[Path] = None,
        key: str = "",
        value: Union[None, list, dict, str, int, float] = None,
        **kwargs,
    ) -> None:
        self.value = value
        super().__init__(name, path, key, **kwargs)

    def _update_simple_value(self, parent: Dict[str, Any]) -> None:
        """
        Update the parent key's value by a simple value.

        :param parent: Parent key of the dict
        :type parent: Dict[str, Any]
        """

        logging.debug('Setting "%s" to "%s"', self.key_name, self.value)
        parent[self.key_name] = self.value

    def _update_list_value(self, parent: Dict[str, Any]) -> None:
        """
        Update the parent key's value which is an array. Depending on the new
        value's type, the exiting list will be extended or the new value will
        be appended to the list.

        :param parent: Parent key of the dict
        :type parent: Dict[str, Any]
        """

        if isinstance(self.value, list):
            logging.debug('Extending "%s" by "%s"', self.key_name, self.value)
            parent[self.key_name].extend(self.value)
        else:
            logging.debug('Appending "%s" to "%s"', self.value, self.key_name)
            parent[self.key_name].append(self.value)

    def _update_dict_value(self, parent: Dict[str, Any]) -> None:
        """
        Update the parent key's value which is a dict.

        :param parent: Parent key of the dict
        :type parent: Dict[str, Any]
        """

        logging.debug('Updating "%s" by "%s"', self.key_name, self.value)
        parent[self.key_name].update(self.value)

    def task(self) -> Path:
        """
        Ensure that the given key has the expected value(s). In case the key cannot
        be found, a ``LookupError`` exception will be raised to stop the execution.

        .. warning::

            Since the value can be anything from ``None`` to a list of lists, and
            rule piping passes the 1st argument (``path``) to the next rule the ``value``
            parameter can not be defined in ``__init__`` before the ``path``. Hence
            the ``value`` parameter must have a default value. The default value is
            set to ``None``, which translates to the following:

            Using the ``DictValueExists`` rule and not assigning value to ``value``
            parameter will set the matching ``key``'s value to `None`` by default in
            the document.

        :raises: ``LookupError`` raised if no key can be renamed or both the new and
                 old keys are in the config file
        :return: Return the input path as an output
        :rtype: Path
        """

        parent = self._get_parent()
        value = parent.get(self.key_name)

        is_list_value = isinstance(value, list)
        is_dict_value = isinstance(value, dict)

        logging.debug('Adding value "%s" to key "%s"', self.value, self.key_name)

        if self.value is None or (not is_list_value and not is_dict_value):
            self._update_simple_value(parent)
        elif is_list_value:
            self._update_list_value(parent)
        elif is_dict_value:
            self._update_dict_value(parent)

        self._write_dump(parent[self.key_name])
        return self.param


class DictValueNotExists(SinglePathDictParsedRule, ABC):
    """
    Ensure that the key has no value given. In case the key cannot be found,
    a ``LookupError`` exception will be raised to stop the execution.

    Compared to ``hammurabi.rules.dictionaries.DictValueExists``, this rule can only
    accept simple value for its ``value`` parameter. No ``list``, ``dict``, or
    ``None`` can be used.

    Based on the key's value's type if the value contains (or equals for simple types)
    value provided in the ``value`` parameter the value is:

    1. Set to None (if the key's value's type is not a dict or list)
    2. Removed from the list (if the key's value's type is a list)
    3. Removed from the dict (if the key's value's type is a dict)

    Example usage:

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar
        >>> from hammurabi.rules.dictionaries import DictValueNotExists
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         DictValueNotExists(
        >>>             name="Remove decommissioned service from dependencies",
        >>>             path=Path("./service.dictionary"),
        >>>             key="development.dependencies",
        >>>             value="service4",
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)
    """

    def __init__(
        self,
        name: str,
        path: Optional[Path] = None,
        key: str = "",
        value: Union[str, int, float] = None,
        **kwargs,
    ) -> None:
        self.value = self.validate(value, required=True)
        super().__init__(name, path, key, **kwargs)

    def task(self) -> Path:
        """
        Ensure that the key has no value given. In case the key cannot be found,
        a ``LookupError`` exception will be raised to stop the execution.

        Based on the key's value's type if the value contains (or equals for simple types)
        value provided in the ``value`` parameter the value is:
        1. Set to None (if the key's value's type is not a dict or list)
        2. Removed from the list (if the key's value's type is a list)
        3. Removed from the dict (if the key's value's type is a dict)

        :return: Return the input path as an output
        :rtype: Path
        """

        parent = self._get_parent()

        value = parent.get(self.key_name)
        value_contains = value and self.value in value

        if self.key_name not in parent:
            return self.param

        write_needed = False
        logging.debug('Removing "%s" from key "%s"', self.value, self.key_name)

        if self.value == value:
            parent[self.key_name] = None
            write_needed = True
        elif isinstance(value, list) and value_contains:
            parent[self.key_name].remove(self.value)
            write_needed = True
        elif isinstance(value, dict) and value_contains:
            del parent[self.key_name][self.value]
            write_needed = True

        if write_needed:
            self._write_dump(parent[self.key_name])

        return self.param



class AbortLawError(Exception):
    """
    Custom exception to make sure that own exception types are
    caught by the Law's execution.
    """


class PreconditionFailedError(Exception):
    """
    Custom exception representing a failed precondition. In case a
    precondition failed, there is no need to raise an error and report
    the rule as a failure. The precondition is for checking that a rule
    should or shouldn't run; not for breaking the execution.
    """



"""
This module contains the definition of Law which is responsible for
the execution of its registered Rules. Every Law can have multiple rules to execute.

In case a rule raises an exception the execution may abort and none of
the remaining rules will be executed neither pipes or children. An abort
can cause an inconsistent state or a dirty git branch. If ``rule_can_abort``
config is set to True, the whole execution of the :class:``hammurabi.pillar.Pillar``
will be aborted and the original exception will be re-raised.
"""

import logging
from typing import Iterable, List, Tuple, Union

from hammurabi.config import config
from hammurabi.exceptions import AbortLawError, PreconditionFailedError
from hammurabi.helpers import full_strip
from hammurabi.mixins import GitMixin
from hammurabi.rules.base import Precondition, Rule


class Law(GitMixin):
    """
    A Law is a collection of Rules which is responsible for the rule execution
    and git committing.

    Example usage:

    .. code-block:: python

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, FileExists
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         FileExists(
        >>>             name="Create pyproject.toml",
        >>>             path=Path("./pyproject.toml")
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)
    """

    def __init__(
        self,
        name: str,
        description: str,
        rules: Iterable[Rule],
        preconditions: Iterable[Precondition] = (),
    ) -> None:
        """
        :param name: Name of the law
        :type name: str

        :param description: Detailed description what kind of rules are included
        :type description: str

        :param rules: List of those rules which should be included in the law
        :type rules: Iterable[Rule]
        """

        super().__init__()

        self.name = name.strip()
        self.description = full_strip(description)
        self.rules: Iterable[Rule] = tuple()
        self.preconditions = preconditions
        self._failed_rules: Tuple[Rule, ...] = tuple()

        for rule in rules:
            self.rules += (rule,)

    def __repr__(self) -> str:
        return (
            "Law("
            f'name="{self.name}", '
            f'description="{self.description}", '
            f"rules={self.rules}, "
            f"preconditions={self.preconditions})"
        )

    def __str__(self) -> str:
        return f"{self.name} law"

    @property
    def passed_rules(self) -> Tuple[Rule, ...]:
        """
        Return the rules which did modifications and not failed.

        :return: Return the passed rules
        :rtype: Tuple[Rule, ...]
        """

        return tuple(r for r in self.rules if r.made_changes)

    @property
    def failed_rules(self) -> Tuple[Rule, ...]:
        """
        Return the rules which did modifications and failed.

        :return: Return the failed rules
        :rtype: Union[Tuple[()], Tuple[Rule]]
        """

        return self._failed_rules

    @property
    def skipped_rules(self) -> Tuple[Rule, ...]:
        """
        Return the rules which neither modified the code nor failed.

        :return: Return the skipped rules
        :rtype: Tuple[Rule, ...]
        """

        def is_skipped(rule) -> bool:
            """
            Return the evaluation if the rule is skipped or not.

            :return: Evaluation if the rule is skipped
            :rtype: bool
            """
            not_passed = rule not in self.passed_rules
            not_failed = rule not in self.failed_rules
            return not_passed and not_failed

        return tuple(r for r in self.rules if is_skipped(r))

    @property
    def documentation(self) -> str:
        """
        Get the name and description of the Law object.

        :return: Return the name and description of the law as its documentation
        :rtype: str
        """

        return f"{self.name}\n{self.description}"

    @property
    def can_proceed(self) -> bool:
        """
        Evaluate if the execution can be continued. If preconditions are set,
        those will be evaluated by this method.

        :return: Return with the result of evaluation
        :rtype: bool

        .. warning::

            :func:`hammurabi.rules.base.Rule.can_proceed` checks the result of
            ``self.preconditions``, which means the preconditions are executed.
            Make sure that you are not doing any modifications within rules used
            as preconditions, otherwise take extra attention for those rules.
        """

        logging.debug('Checking if "%s" can proceed with execution', self.name)
        return all([condition.execute() for condition in self.preconditions])

    def get_execution_order(self) -> List[Union[Rule, Precondition]]:
        """
        Get the execution order of the registered rules. The order will
        contain the pipes and children as well.

        This helper function is useful in debugging and information
        gathering.

        :return: Return the execution order of the rules
        :rtype: List[Rule]
        """

        order: List[Union[Rule, Precondition]] = list()

        for rule in self.rules:
            order.extend(rule.get_execution_order())

        return order

    def commit(self) -> None:
        """
        Commit the changes made by registered rules and add a
        meaningful commit message.

        Example commit message:

        .. code-block:: text

            Migrate to next generation project template
            * Create pyproject.toml
            * Add meta info from setup.py to pyproject.toml
            * Add existing dependencies
            * Remove requirements.txt
            * Remove setup.py
        """

        if not self.passed_rules:
            return

        order = self.get_execution_order()
        rules = [f"* {rule.name}" for rule in order if rule.made_changes]
        rules_commit_message = "\n".join(rules)
        self.git_commit(f"{self.documentation}\n\n{rules_commit_message}")

    @staticmethod
    def __execute_rule_chain(rule: Rule) -> None:
        """
        Execute the given rule. In case of an exception, the execution of rules
        will continue except the failing one. The failed rule's pipe and children
        will not be executed.

        :param rule: A registered rule
        :type rule: Rule

        :raises: ``AbortLawError``
        """

        try:
            rule.execute()
        except PreconditionFailedError:
            logging.warning(
                'Cancelling execution of "%s", the preconditions are not fulfilled',
                rule.name,
            )
        except Exception as exc:  # pylint: disable=broad-except
            logging.error('Execution of "%s" is aborted: %s', rule.name, str(exc))

            chained_rules = filter(
                lambda r: isinstance(r, Rule), rule.get_rule_chain(rule)
            )

            for chain in chained_rules:
                logging.error('Due to errors "%s" is aborted', chain.name)

            raise AbortLawError(str(exc)) from exc

    def __execute_rule(self, rule: Rule) -> None:
        """
        Execute a rule registered for the law. In case of an exception the
        exception will be re-raised.
        """

        try:
            self.__execute_rule_chain(rule)
        except AbortLawError as exc:
            logging.error(str(exc))
            self._failed_rules += (rule,)

            if config.settings.rule_can_abort:
                raise exc

    def enforce(self) -> None:
        """
        Execute all registered rule. If ``rule_can_abort`` config option
        is set to ``True``, all the rules will be aborted and an exception
        will be raised.

        When the whole execution chain is finished, the changes will be
        committed except the failed ones.

        .. note::

            Failed rules and their chain (excluding prerequisites) will be added
            to the pull request description.

        :raises: ``AbortLawError``
        """

        if not self.can_proceed:
            logging.warning(
                'Cancelling execution of "%s", the preconditions are not fulfilled',
                self.name,
            )
            return

        logging.info('Executing law "%s"', self.name)

        for rule in self.rules:
            self.__execute_rule(rule)

        logging.info('Committing changes made by "%s"', self.name)
        self.commit()



"""
Mixins module contains helpers for both laws and rules. Usually this file
will contain Git commands related helpers. Also, this module contains the
extensions for several online git based VCS.
"""

import logging
from pathlib import Path
from typing import Iterable, List, Optional, Union

from github3.pulls import ShortPullRequest  # type: ignore
from github3.repos.repo import Repository  # type: ignore
from github3.structs import GitHubIterator  # type: ignore

from hammurabi.config import config
from hammurabi.preconditions.base import Precondition
from hammurabi.rules.base import Rule


class GitMixin:
    """
    Simple mixin which contains all the common git commands which are needed
    to push a change to an online VCS like GitHub or GitLab. This mixin could
    be used by :class:`hammurabi.law.Law`s, :class:`hammurabi.rules.base` or
    any rules which can make modifications during its execution.
    """

    @staticmethod
    def __can_proceed() -> bool:
        """
        Determine if the next change can be done or not. For git related
        operations it is extremely important to abort or skip an execution
        if not needed.

        :return: Returns True if the next changes should be done
        :rtype: bool
        """

        is_dirty = config.repo and config.repo.is_dirty(untracked_files=True)
        return is_dirty and not config.settings.dry_run

    @staticmethod
    def checkout_branch() -> None:
        """
        Perform a simple git checkout, to not pollute the default branch and
        use that branch for the pull request later. The branch name can be
        changed in the config by setting the ``git_branch_name`` config option.

        The following command is executed:

        .. code-block:: shell

            git checkout -b <branch name>
        """

        if config.repo and not config.settings.dry_run:
            branch = config.settings.git_branch_name
            logging.info('Checkout branch "%s"', branch)
            config.repo.git.checkout("HEAD", B=branch)  # pylint: disable=no-member
            config.repo.git.pull("origin", branch)

    def git_add(self, param: Path) -> None:
        """
        Add file contents to the index.

        :param param: Path to add to the index
        :type param: Path

        The following command is executed:

        .. code-block:: shell

            git add <path>
        """

        if self.__can_proceed():
            logging.debug('Git add "%s"', str(param))
            config.repo.git.add(str(param))  # pylint: disable=no-member
            self.made_changes = True

    def git_remove(self, param: Path) -> None:
        """
        Remove files from the working tree and from the index.

        :param param: Path to remove from the working tree and the index
        :type param: Path

        The following command is executed:

        .. code-block:: shell

            git rm <path>
        """

        if self.__can_proceed():
            logging.debug('Git remove "%s"', str(param))
            config.repo.index.remove(
                (str(param),), ignore_unmatch=True, r=True
            )  # pylint: disable=no-member
            self.made_changes = True

    def git_commit(self, message: str) -> None:
        """
        Commit the changes on the checked out branch.

        :param message: Git commit message
        :type message: str

        The following command is executed:

        .. code-block:: shell

            git commit -m "<commit message>"
        """

        if self.__can_proceed():
            config.repo.index.commit(message)  # pylint: disable=no-member

    @staticmethod
    def push_changes() -> None:
        """
        Push the changes with the given branch set by ``git_branch_name``
        config option to the remote origin.

        The following command is executed:

        .. code-block:: shell

            git push origin <branch name>
        """

        if config.repo and not config.settings.dry_run:
            branch: str = config.settings.git_branch_name
            logging.info("Pushing changes to %s", branch)
            config.repo.remotes.origin.push(branch)  # pylint: disable=no-member


class PullRequestHelperMixin:  # pylint: disable=too-few-public-methods
    """
    Give helper classes for pull request related operations
    """

    @staticmethod
    def __get_chained_rules(
        target: Rule, chain: List[Union[Rule, Precondition]]
    ) -> Iterable[Rule]:
        """
        Return all the chained rules excluding the root rule.

        :param target: The root Rule
        :type target: Rule

        :param chain: The whole chain
        :type chain: Iterable[Rule]

        :return: The filtered list of chained rules
        :rtype: Iterable[Rule]
        """

        rules = filter(lambda i: isinstance(i, Rule), chain)
        return filter(lambda r: r != target, rules)  # type: ignore

    def __get_rules_body(self, rules: Iterable[Rule]) -> List[str]:
        """
        Generate the PR body's executed rules section for the root
        and chained rules excluding their preconditions.

        :param rules: List of passing of failed rules
        :type rules: Iterable[Rule]

        :return: Body of the PR section in a list format
        :rtype: List[str]
        """

        body: List[str] = list()

        for rule in rules:
            body.append(f"* {rule.name}")

            for chain in self.__get_chained_rules(rule, rule.get_rule_chain(rule)):
                body.append(f"  * {chain.name}")

        return body

    def generate_pull_request_body(self, pillar) -> str:
        """
        Generate the body of the pull request based on the registered laws and rules.
        The pull request body is markdown formatted.

        :param pillar: Pillar configuration
        :type pillar: :class:`hammurabi.pillar.Pillar`

        :return: Returns the generated pull request description
        :rtype: str
        """

        #  NOTE: The parameter type can not be hinted, because of circular import,
        #  we must fix this in the future releases.

        logging.debug("Generating pull request body")

        body: List[str] = [
            "## Description",
            "Below you can find the executed laws and information about them.",
        ]

        # Filter only those laws which has other rules than skipped ones.
        actionable_laws = filter(lambda l: l.passed_rules + l.failed_rules, pillar.laws)

        for law in actionable_laws:
            body.append(f"\n### {law.name}")
            body.append(law.description)

            if law.passed_rules:
                body.append("\n<details>")
                body.append("<summary>Passed rules</summary>\n")
                body.extend(self.__get_rules_body(law.passed_rules))
                body.append("</details>")

            if law.failed_rules:
                body.append("\n<details open>")
                body.append("<summary>Failed rules (manual fix needed)</summary>\n")
                body.extend(self.__get_rules_body(law.failed_rules))
                body.append("</details>")

        return "\n".join(body)


# TODO: Move to a separate file as part of a new package if other VCS is supported
class GitHubMixin(GitMixin, PullRequestHelperMixin):
    """
    Extending :class:`hammurabi.mixins.GitMixin` to be able to open pull requests
    on GitHub after changes are pushed to remote.
    """

    def create_pull_request(self) -> Optional[str]:
        """
        Create a PR on GitHub after the changes are pushed to remote. The pull
        request details (repository, branch) are set by the project
        configuration. The mapping of the details and configs:

        +------------+--------------------------------------+
        | Detail     | Configuration                        |
        +============+======================================+
        | repo       | repository (owner/repository format) |
        +------------+--------------------------------------+
        | base       | git_base_name                        |
        +------------+--------------------------------------+
        | branch     | git_branch_name                      |
        +------------+--------------------------------------+

        :return: Return the open (and updated) or opened PR's url
        :rtype: Optional[str]
        """

        if not config.github:
            raise RuntimeError(
                "The GitHub client is not initialized properly. Make sure that "
                "you set the GITHUB_TOKEN or HAMMURABI_GITHUB_TOKEN before execution. "
            )

        if config.repo and not config.settings.dry_run:
            owner, repository = config.settings.repository.split("/")
            github_repo: Repository = config.github.repository(owner, repository)

            logging.info("Checking for opened pull request")

            pull_request_iterator: GitHubIterator[
                ShortPullRequest
            ] = github_repo.pull_requests(
                state="open",
                head=config.settings.git_branch_name,
                base=config.settings.git_base_name,
                number=1,  # Maximum number of PRs
            )

            opened_pull_request: Optional[ShortPullRequest] = next(
                pull_request_iterator, None
            )

            if not opened_pull_request:
                description = self.generate_pull_request_body(config.settings.pillar)

                logging.info("Opening pull request")
                response: ShortPullRequest = github_repo.create_pull(
                    title="[hammurabi] Update to match the latest baseline",
                    base=config.settings.git_base_name,
                    head=config.settings.git_branch_name,
                    body=description,
                )

                return response.url

            # Return the last known PR url, it should be one
            # anyways, so it is not an issue
            pull_request_url = opened_pull_request.url
            logging.info("Found open pull request %s", pull_request_url)
            return pull_request_url

        # Although this return could be skipped, it is more
        # explicit to have it here
        return None



"""
Templates module contains rules which are capable to create a new
file based on a Jinja2 template by rendering it.
"""

import logging
from pathlib import Path
from typing import Any, Dict, Optional

from jinja2 import Template

from hammurabi.rules.common import SinglePathRule


class TemplateRendered(SinglePathRule):
    """
    Render a file from a Jinja2 template. In case the destination
    file not exists, this rule will create it, otherwise the file will
    be overridden.

    Example usage:

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, TemplateRendered
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         TemplateRendered(
        >>>             name="Create gunicorn config from template",
        >>>             template=Path("/tmp/templates/gunicorn.conf.py"),
        >>>             destination=Path("./gunicorn.conf.py"),
        >>>             context={
        >>>                 "keepalive": 65
        >>>             },
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)
    """

    def __init__(
        self,
        name: str,
        template: Optional[Path] = None,
        destination: Optional[Path] = None,
        context: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> None:
        self.destination: Path = self.validate(destination, required=True)
        self.context: Dict[str, Any] = context or dict()
        super().__init__(name, template, **kwargs)

    def post_task_hook(self):
        """
        Add the destination and not the original path.
        """

        self.git_add(self.destination)

    def task(self) -> Path:
        """
        Render a file from a Jinja2 template. In case the destination
        file not exists, this rule will create it, otherwise the file will
        be overridden.

        :return: Returns the path of the rendered file
        :rtype: Path
        """

        logging.debug('Rendering template "%s"', str(self.param))
        rendered = Template(self.param.read_text()).render(self.context)
        self.destination.write_text(rendered)

        return self.destination



"""
Files module contains file specific manipulation rules. Please note that
those rules which can be used for files and directories are located in
other modules like :mod:`hammurabi.rules.operations` or
:mod:`hammurabi.rules.attributes`.
"""

import logging
from pathlib import Path
from typing import Iterable

from hammurabi.rules.common import MultiplePathRule, SinglePathRule


class FileExists(SinglePathRule):
    """
    Ensure that a file exists. If the file does not exists,
    make sure the file is created.

    Example usage:

    .. code-block:: python

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, FileExists
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         FileExists(
        >>>             name="Create service descriptor",
        >>>             path=Path("./service.yaml")
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)
    """

    def task(self) -> Path:
        """
        If the target file not exists, create the file to make sure we
        can manipulate it.

        :return: The created/existing file's path
        :rtype: Path
        """

        logging.debug('Creating file "%s" if not exists', str(self.param))
        self.param.touch()

        return self.param


class FilesExist(MultiplePathRule):
    """
    Ensure that all files exists. If the files does not exists,
    make sure the files are created.

    Example usage:

    .. code-block:: python

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, FilesExist
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         FilesExist(
        >>>             name="Create test files",
        >>>             paths=[
        >>>                 Path("./file_1"),
        >>>                 Path("./file_2"),
        >>>                 Path("./file_3"),
        >>>             ]
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)
    """

    def task(self) -> Iterable[Path]:
        """
        If the target files not exist, create the files to make sure we
        can manipulate them.

        :return: The created/existing files' path
        :rtype: Iterable[Path]
        """

        for path in self.param:
            logging.debug('Creating file "%s" if not exists', str(path))
            path.touch()

        return self.param


class FileNotExists(SinglePathRule):
    """
    Ensure that the given file does not exists. If the file exists
    remove it, otherwise do nothing and return the original path.

    Example usage:

    .. code-block:: python

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, FileNotExists
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         FileNotExists(
        >>>             name="Remove unused file",
        >>>             path=Path("./debug.yaml")
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)
    """

    def post_task_hook(self):
        """
        Remove the given file from git index.
        """

        self.git_remove(self.param)

    def task(self) -> Path:
        """
        Remove the given file if exists, otherwise do nothing and
        return the original path.

        :return: Return the removed file's path
        :rtype: Path
        """

        if self.param.exists():
            logging.debug('Removing "%s"', str(self.param))
            self.param.unlink()

        return self.param


class FilesNotExist(MultiplePathRule):
    """
    Ensure that the given files does not exist. If the files exist
    remove them, otherwise do nothing and return the original paths.

    Example usage:

    .. code-block:: python

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, FilesNotExist
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         FilesNotExist(
        >>>             name="Remove several files",
        >>>             paths=[
        >>>                 Path("./file_1"),
        >>>                 Path("./file_2"),
        >>>                 Path("./file_3"),
        >>>             ]
        >>>         ),
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)
    """

    def post_task_hook(self):
        """
        Remove the given files from git index.
        """

        for path in self.param:
            self.git_remove(path)

    def task(self) -> Iterable[Path]:
        """
        Remove all existing files.

        :return: Return the removed files' paths
        :rtype: Iterable[Path]
        """

        for path in self.param:
            if path.exists():
                logging.debug('Removing "%s"', str(path))
                path.unlink()

        return self.param


class FileEmptied(SinglePathRule):
    """
    Remove the content of the given file, but keep the file. Please note the
    difference between emptying a file and recreating it. The latter
    results in lost ACLs, permissions and modes.

    Example usage:

    .. code-block:: python

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, FileEmptied
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         FileEmptied(
        >>>             name="Empty the check log file",
        >>>             path=Path("/var/log/service/check.log")
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)
    """

    def task(self) -> Path:
        """
        Remove the content of the given file. If the file does not exists
        this rule will create the file without content.

        :return: Return the emptied/created file's path
        :rtype: Path
        """

        logging.debug('Emptying "%s"', str(self.param))
        self.param.write_text("")

        return self.param



"""
This module adds JSON file support. JSON module is an extension for text rules
tailor made for .json files. The main difference lies in the way it works.
First, the .json file is parsed, then the modifications are made on the
already parsed file.
"""

from abc import abstractmethod
import json
from pathlib import Path
from typing import Any, Optional

from hammurabi.rules.dictionaries import (
    DictKeyExists,
    DictKeyNotExists,
    DictKeyRenamed,
    DictValueExists,
    DictValueNotExists,
    SinglePathDictParsedRule,
)


class SingleJSONFileRule(SinglePathDictParsedRule):
    """
    Extend :class:`hammurabi.rules.dictionaries.SinglePathDictParsedRule`
    to handle parsed content manipulations on a single JSON file.
    """

    def __init__(
        self, name: str, path: Optional[Path] = None, key: str = "", **kwargs
    ) -> None:
        super().__init__(name, path, key, loader=json.loads, **kwargs)

    def _write_dump(self, data: Any, delete: bool = False) -> None:
        """
        Helper function to write the dump into file.

        :param data: The modified data
        :type data: :class:``hammurabi.rules.mixins.Any`

        :param delete: Indicate if the key should be deleted
        :type delete: bool
        """

        self.param.write_text(
            json.dumps(
                self.set_by_selector(self.loaded_data, self.split_key, data, delete)
            )
        )

    @abstractmethod
    def task(self) -> Path:
        """
        Abstract method representing how a :func:`hammurabi.rules.base.Rule.task`
        must be parameterized. Any difference in the parameters will result in
        pylint/mypy errors.

        For more details please check :func:`hammurabi.rules.base.Rule.task`.
        """


class JSONKeyExists(DictKeyExists, SingleJSONFileRule):
    """
    Ensure that the given key exists. If needed, the rule will create a key with the
    given name, and optionally the specified value. In case the value is set, the value
    will be assigned to the key. If no value is set, the key will be created with an empty
    value.

    Example usage:

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, JSONKeyExists
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         JSONKeyExists(
        >>>             name="Ensure service descriptor has stack",
        >>>             path=Path("./service.json"),
        >>>             key="stack",
        >>>             value="my-awesome-stack",
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)

    .. warning::

        Compared to :mod:`hammurabi.rules.text.LineExists`, this rule is NOT able to add a
        key before or after a target.
    """


class JSONKeyNotExists(DictKeyNotExists, SingleJSONFileRule):
    """
    Ensure that the given key not exists. If needed, the rule will remove a key with the
    given name, including its value.

    Example usage:

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, JSONKeyNotExists
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         JSONKeyNotExists(
        >>>             name="Ensure outdated_key is removed",
        >>>             path=Path("./service.json"),
        >>>             key="outdated_key",
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)
    """


class JSONKeyRenamed(DictKeyRenamed, SingleJSONFileRule):
    """
    Ensure that the given key is renamed. In case the key can not be found,
    a ``LookupError`` exception will be raised to stop the execution. The
    execution must be stopped at this point, because if other rules depending
    on the rename they will fail otherwise.

    Example usage:

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, JSONKeyRenamed
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         JSONKeyRenamed(
        >>>             name="Ensure service descriptor has dependencies",
        >>>             path=Path("./service.json"),
        >>>             key="development.depends_on",
        >>>             value="dependencies",
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)
    """


class JSONValueExists(DictValueExists, SingleJSONFileRule):
    """
    Ensure that the given key has the expected value(s). In case the key cannot
    be found, a ``LookupError`` exception will be raised to stop the execution.

    This rule is special in the way that the value can be almost anything. For
    more information please read the warning below.

    Example usage:

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, JSONValueExists
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         JSONValueExists(
        >>>             name="Ensure service descriptor has dependencies",
        >>>             path=Path("./service.json"),
        >>>             key="development.dependencies",
        >>>             value=["service1", "service2", "service3"],
        >>>         ),
        >>>         # Or
        >>>         JSONValueExists(
        >>>             name="Add infra alerting to existing alerting components",
        >>>             path=Path("./service.json"),
        >>>             key="development.alerting",
        >>>             value={"infra": "#slack-channel-2"},
        >>>         ),
        >>>         # Or
        >>>         JSONValueExists(
        >>>             name="Add support info",
        >>>             path=Path("./service.json"),
        >>>             key="development.supported",
        >>>             value=True,
        >>>         ),
        >>>         # Or even
        >>>         JSONValueExists(
        >>>             name="Make sure that no development branch is set",
        >>>             path=Path("./service.json"),
        >>>             key="development.branch",
        >>>             value=None,
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)

    .. warning::

        Since the value can be anything from ``None`` to a list of lists, and
        rule piping passes the 1st argument (``path``) to the next rule the ``value``
        parameter can not be defined in ``__init__`` before the ``path``. Hence
        the ``value`` parameter must have a default value. The default value is
        set to ``None``, which translates to the following:

        Using the ``JSONValueExists`` rule and not assigning value to ``value``
        parameter will set the matching ``key``'s value to `None`` by default in
        the document.
    """


class JSONValueNotExists(DictValueNotExists, SingleJSONFileRule):
    """
    Ensure that the key has no value given. In case the key cannot be found,
    a ``LookupError`` exception will be raised to stop the execution.

    Compared to ``hammurabi.rules.json.JSONValueExists``, this rule can only
    accept simple value for its ``value`` parameter. No ``list``, ``dict``, or
    ``None`` can be used.

    Based on the key's value's type if the value contains (or equals for simple types)
    value provided in the ``value`` parameter the value is:

    1. Set to None (if the key's value's type is not a dict or list)
    2. Removed from the list (if the key's value's type is a list)
    3. Removed from the dict (if the key's value's type is a dict)

    Example usage:

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, JSONValueNotExists
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         JSONValueNotExists(
        >>>             name="Remove decommissioned service from dependencies",
        >>>             path=Path("./service.json"),
        >>>             key="development.dependencies",
        >>>             value="service4",
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)
    """



# pylint: disable=too-few-public-methods

"""
This module contains the definition of Reporters which is responsible for
exposing the execution results in several formats.
"""
from abc import ABC, abstractmethod
from datetime import datetime
import logging
from pathlib import Path
from typing import Any, Iterable, List

from pydantic import BaseModel  # pylint: disable=no-name-in-module

from hammurabi.config import config
from hammurabi.law import Law
from hammurabi.rules.base import Rule


class LawItem(BaseModel):
    """
    LawItem represents the basic summary of a low attached to a rule.
    """

    name: str
    description: str


class RuleItem(BaseModel):
    """
    RuleItem represents the registered rule and its status.

    The rule (as normally) has the status of the execution which can be
    passed, failed or skipped.
    """

    name: str
    law: LawItem


class AdditionalData(BaseModel):
    """
    Additional data which may not be set for every execution.
    """

    # Add the execution start and finish datetime
    # to the report for statistic purposes
    started: str = datetime.min.isoformat()
    finished: str = datetime.min.isoformat()

    # When PR is not created, still do the reporting
    pull_request_url: str = ""


class Report(BaseModel):
    """
    The report object which contains all the necessary and optional
    data for the report will be generated.
    """

    passed: List[RuleItem] = list()
    failed: List[RuleItem] = list()
    skipped: List[RuleItem] = list()
    additional_data: AdditionalData = AdditionalData()


class Reporter(ABC):
    """
    Abstract class which describes the bare minimum and helper functions for Reporters.
    A reporter can generate different outputs from the results of the execution. Also,
    reporters can be extended by additional data which may not contain data for every
    execution like GitHub pull request url. The report file's name set by ``report_name``
    config parameter.

    .. note::

        Reporters measures the execution time for the complete execution from checking
        out the git branch until the pull request creation finished. Although the
        completion time is measured, it is not detailed for the rules. At this moment
        measuring execution time of rules is not planned.

    Example usage:

    .. code-block:: python

        >>> from hammurabi.reporters.base import Reporter
        >>>
        >>>
        >>> class JSONReporter(Reporter):
        >>>     def report(self) -> str:
        >>>         return self._get_report().json()

    :param laws: Iterable Law objects which will be included to the report
    :type laws: Iterable[Law]
    """

    def __init__(self, laws: List[Law]) -> None:
        self.laws = laws
        self.additional_data: AdditionalData = AdditionalData()
        self.report_path: Path = config.settings.report_name

    @staticmethod
    def __get_rule_item_from_rule(law: Law, rule: Rule, **kwargs) -> RuleItem:
        """
        Get a rule item from a combination of a Rule and a Law.

        :param law: The parent law of the rule
        :type law: ``hammurabi.law.Law``

        :param rule: The rule which should be transformed
        :type rule: ``hammurabi.rules.base.Rule``

        :return: Return the transformed RuleItem from the given law and rule
        :rtype: RuleItem
        """

        logging.info('Collecting report data for "%s"', str(rule))

        return RuleItem(
            name=rule.name,
            law=LawItem(name=law.name, description=law.description),
            **kwargs
        )

    def __get_rule_items(
        self, law: Law, rules: Iterable[Rule], **kwargs
    ) -> List[RuleItem]:
        """
        Get all the passed rule items for a law.

        :param law: The parent law of the rule
        :type law: ``hammurabi.law.Law``

        :return: Return all the passed rules transformed to ``RuleItem``s
        :rtype: List[RuleItem]
        """

        logging.info('Collecting report data for "%s"', str(law))

        items: List[RuleItem] = list()

        for rule in rules:
            items += (self.__get_rule_item_from_rule(law, rule, **kwargs),)

        return items

    def _get_report(self) -> Report:
        """
        Get and prepare the report for actual reporting.

        :return: Return the assembled report object
        :rtype: ``hammurabi.reporters.base.Report``
        """

        logging.info("Generating execution report")

        report = Report(additional_data=self.additional_data)

        for law in self.laws:
            report.passed += self.__get_rule_items(law, law.passed_rules)
            report.failed += self.__get_rule_items(law, law.failed_rules)
            report.skipped += self.__get_rule_items(law, law.skipped_rules)

        logging.info("Execution report generated")

        return report

    @abstractmethod
    def report(self) -> Any:
        """
        Do the actual reporting based on the report assembled.
        """



"""
This module contains the definition of Preconditions which describes what to do with
the received parameter and does the necessary changes. The preconditions are used to
enable developers skipping or enabling rules based on a set of conditions.

.. warning::

    The precondition is for checking that a rule should or shouldn't run, not for
    breaking/aborting the execution. To indicate a precondition failure as an error
    in the logs, create a precondition which raises an exception if the requirements
    doesn't match.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
import logging
from typing import Any, Optional

from hammurabi.rules.abstract import AbstractRule


class Precondition(AbstractRule, ABC):
    """
    This class which describes the bare minimum and helper functions for Preconditions.
    A precondition defines what and how should be checked/validated before executing a Rule.
    Since preconditions are special rules, all the functions available what can be used for
    :class:`hammurabi.rules.base.AbstractRule`.

    As said, preconditions are special from different angles. While this is not true for
    Rules, Preconditions will always have a name, hence giving a name to a Precondition is not
    necessary. In case no name given to a precondition, the name will be the name of the class
    and " precondition" suffix.

    Example usage:

    .. code-block:: python

        >>> import logging
        >>> from typing import Optional
        >>> from pathlib import Path
        >>> from hammurabi import Precondition
        >>>
        >>> class IsFileExists(Precondition):
        >>>     def __init__(self, path: Optional[Path] = None, **kwargs) -> None:
        >>>         super().__init__(param=path, **kwargs)
        >>>
        >>>     def task(self) -> bool:
        >>>         return self.param and self.param.exists()

    :param name: Name of the rule which will be used for printing
    :type name: Optional[str]

    :param param: Input parameter of the rule will be used as ``self.param``
    :type param: Any

    .. note:

        Since ``Precondition`` inherits from ``Rule``, the parameter after the name of the
        precondition will be used for ``self.param``. This can be handy for interacting
        with input parameters.

    .. warning:

        Although ``Precondition`` inherits from ``Rule``, the pipe and children execution
        is intentionally not implemented.
    """

    def __init__(self, name: Optional[str] = None, param: Optional[Any] = None) -> None:
        name = name or f"{self.__class__.__name__} precondition"
        super().__init__(name, param)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(name="{self.name}", param="{self.param}")'

    def __str__(self) -> str:
        if self.name.endswith("precondition"):
            return self.name

        return f"{self.name} precondition"

    @abstractmethod
    def task(self) -> bool:
        """
        Abstract method representing how a :func:`hammurabi.rules.base.Precondition.task`
        must be parameterized. Any difference in the parameters or return type will result
        in pylint/mypy errors.

        To be able to use the power of ``pipe`` and ``children``, return
        something which can be generally used for other rules as in input.

        :return: Returns an output which can be used as an input for other rules
        :rtype: Any (usually same as `self.param`'s type)
        """

    def execute(self) -> bool:
        """
        Execute the precondition.

        :raise: ``AssertionError``
        :return: None
        """

        logging.info('Running task for "%s"', self.name)

        self.pre_task_hook()
        result = self.task()
        self.post_task_hook()

        return result



"""
Pillar module is responsible for handling the whole execution chain including
executing the registered laws, pushing the changes to the VCS and creating a
pull request. All the laws registered to the pillar will be executed in the
order of the registration.
"""


from datetime import datetime
from typing import List, Type

from hammurabi.law import Law
from hammurabi.mixins import GitHubMixin
from hammurabi.reporters.base import Reporter
from hammurabi.reporters.json import JSONReporter
from hammurabi.rules.base import Rule


class Pillar(GitHubMixin):
    """
    Pillar is responsible for the execution of the chain of laws and rules.

    All the registered laws and rules can be retrieved using the ``laws`` and
    ``rules`` properties, or if necessary single laws and rules can be accessed
    using the resource's name as a parameter for ``get_law`` or ``get_rule``
    methods.

    As a final step, pillar will prepare its ``reporter`` for report generation.
    For more information about reporters, check :class:`hammurabi.reporters.base.Reporter`
    and :class:`hammurabi.reporters.json.JSONReporter`.

    :param reporter_class: The reporter class used for generating the reports
    :type reporter_class: Type[Reporter]
    """

    def __init__(self, reporter_class: Type[Reporter] = JSONReporter) -> None:
        self.__laws: List[Law] = list()

        self.reporter: Reporter = reporter_class(list())

    @property
    def laws(self) -> List[Law]:
        """
        Return the registered laws in order of the registration.
        """

        return self.__laws

    @property
    def rules(self) -> List[Rule]:
        """
        Return all the registered laws' rules.
        """

        return [rule for law in self.laws for rule in law.rules]

    def get_law(self, name: str) -> Law:
        """
        Get a law by its name. In case of no Laws are registered or
        the law can not be found by its name, a ``StopIteration``
        exception will be raised.

        :param name: Name of the law which will be used for the lookup
        :type name: str

        :raises: ``StopIteration`` exception if Law not found
        :return: Return the searched law
        :rtype: :class:`hammurabi.law.Law`
        """

        return next(filter(lambda l: l.name == name, self.laws))

    def get_rule(self, name: str) -> Rule:
        """
        Get a registered rule (and its pipe/children) by the rule's name.

        This helper function is useful in debugging and information
        gathering.

        :param name: Name of the rule which will be used for the lookup
        :type name: str

        :raises: ``StopIteration`` exception if Rule not found
        :return: Return the rule in case of a match for the name
        :rtype: Rule
        """

        return next(filter(lambda r: r.name == name, self.rules))

    def register(self, law: Law):
        """
        Register the given Law to the Pillar. The order of the registration
        does not matter. The laws should never depend on each other.

        :param law: Initialized Law which should be registered
        :type law: ``hammurabi.law.Law``

        Example usage:

        .. code-block:: python

            >>> from pathlib import Path
            >>> from hammurabi import Law, Pillar, FileExists
            >>>
            >>> example_law = Law(
            >>>     name="Name of the law",
            >>>     description="Well detailed description what this law does.",
            >>>     rules=(
            >>>         FileExists(
            >>>             name="Create pyproject.toml",
            >>>             path=Path("./pyproject.toml")
            >>>         ),
            >>>     )
            >>> )
            >>>
            >>> pillar = Pillar()
            >>> pillar.register(example_law)

        .. warning::

            The laws should never depend on each other, because the execution
            may not happen in the same order the laws were registered. Instead,
            organize the depending rules in one law to resolve any dependency
            conflicts.
        """

        self.__laws.append(law)
        self.reporter.laws = self.__laws

    def enforce(self):
        """
        Run all the registered laws and rules one by one. This method is responsible
        for executing the registered laws, push changes to the git origin and open
        the pull request.

        This method glues together the lower level components and makes sure that the
        execution of laws and rules can not be called more than once at the same time
        for a target.
        """

        self.reporter.additional_data.started = datetime.now().isoformat()
        self.checkout_branch()

        for law in self.laws:
            law.enforce()

        self.push_changes()
        pull_request_url = self.create_pull_request()

        self.reporter.additional_data.finished = datetime.now().isoformat()
        self.reporter.additional_data.pull_request_url = pull_request_url



from abc import abstractmethod
from pathlib import Path
from typing import Any, Iterable, Optional

from hammurabi.mixins import GitMixin
from hammurabi.rules.base import Rule


class SinglePathRule(Rule, GitMixin):
    """
    Abstract class which extends :class:`hammurabi.rules.base.Rule` to handle operations on a
    single directory.
    """

    def __init__(self, name: str, path: Optional[Path] = None, **kwargs) -> None:
        super().__init__(name, path, **kwargs)

    def post_task_hook(self):
        self.git_add(self.param)

    @abstractmethod
    def task(self) -> Any:
        """
        Abstract method representing how a :func:`hammurabi.rules.base.Rule.task`
        must be parameterized. Any difference in the parameters will result in
        pylint/mypy errors.

        For more details please check :func:`hammurabi.rules.base.Rule.task`.
        """


class MultiplePathRule(Rule, GitMixin):
    """
    Abstract class which extends :class:`hammurabi.rules.base.Rule` to handle operations on
    multiple files.
    """

    def __init__(
        self, name: str, paths: Optional[Iterable[Path]] = (), **kwargs
    ) -> None:
        super().__init__(name, paths, **kwargs)

    def post_task_hook(self):
        for path in self.param:
            self.git_add(path)

    @abstractmethod
    def task(self) -> Any:
        """
        Abstract method representing how a :func:`hammurabi.rules.base.Rule.task`
        must be parameterized. Any difference in the parameters will result in
        pylint/mypy errors.

        For more details please check :func:`hammurabi.rules.base.Rule.task`.
        """



"""
Operations module contains common file/directory operation which can be
handy when need to move, rename or copy files.
"""

import logging
from pathlib import Path
import shutil
from typing import Optional

from hammurabi.rules.common import SinglePathRule


class Moved(SinglePathRule):
    """
    Move a file or directory from "A" to "B".

    Example usage:

    .. code-block:: python

            >>> from pathlib import Path
            >>> from hammurabi import Law, Pillar, Moved
            >>>
            >>> example_law = Law(
            >>>     name="Name of the law",
            >>>     description="Well detailed description what this law does.",
            >>>     rules=(
            >>>         Moved(
            >>>             name="Move pyproject.toml to its place",
            >>>             path=Path("/tmp/generated/pyproject.toml.template"),
            >>>             destination=Path("./pyproject.toml"),  # Notice the rename!
            >>>         ),
            >>>     )
            >>> )
            >>>
            >>> pillar = Pillar()
            >>> pillar.register(example_law)
    """

    def __init__(
        self,
        name: str,
        path: Optional[Path] = None,
        destination: Optional[Path] = None,
        **kwargs,
    ) -> None:
        self.destination = self.validate(destination, required=True)
        super().__init__(name, path, **kwargs)

    def post_task_hook(self):
        """
        Add both the new and old git objects.
        """

        self.git_remove(self.param)
        self.git_add(self.destination)

    def task(self) -> Path:
        """
        Move the given path to the destination. In case the file got a
        new name when destination is provided, the file/directory will
        be moved to its new place with its new name.

        :return: Returns the new destination of the file/directory
        :rtype: Path
        """

        logging.debug('Moving "%s" to "%s"', str(self.param), str(self.destination))
        shutil.move(self.param, self.destination)

        return self.destination


class Renamed(Moved):
    """
    This rule is a shortcut for :class:`hammurabi.rules.operations.Moved`.
    Instead of destination path a new name is required.

    Example usage:

    .. code-block:: python

            >>> from pathlib import Path
            >>> from hammurabi import Law, Pillar, Renamed
            >>>
            >>> example_law = Law(
            >>>     name="Name of the law",
            >>>     description="Well detailed description what this law does.",
            >>>     rules=(
            >>>         Renamed(
            >>>             name="Rename pyproject.toml.bkp",
            >>>             path=Path("/tmp/generated/pyproject.toml.bkp"),
            >>>             new_name="pyproject.toml",
            >>>         ),
            >>>     )
            >>> )
            >>>
            >>> pillar = Pillar()
            >>> pillar.register(example_law)
    """

    def __init__(
        self,
        name: str,
        path: Optional[Path] = None,
        new_name: Optional[str] = None,
        **kwargs,
    ) -> None:
        path_name: str = self.validate(new_name, required=True)
        destination = Path((path or self.param).parent, path_name)
        super().__init__(name, path, destination, **kwargs)


class Copied(SinglePathRule):
    """
    Ensure that the given file or directory is copied to the new path.

    Example usage:

    .. code-block:: python

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, Copied
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         Copied(
        >>>             name="Create backup file",
        >>>             path=Path("./service.yaml"),
        >>>             destination=Path("./service.bkp.yaml")
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)
    """

    def __init__(
        self,
        name: str,
        path: Optional[Path] = None,
        destination: Optional[Path] = None,
        **kwargs,
    ) -> None:
        self.destination = self.validate(destination, required=True)
        super().__init__(name, path, **kwargs)

    def post_task_hook(self):
        """
        Add the destination and not the original path.
        """

        self.git_add(self.destination)

    def task(self) -> Path:
        """
        Copy the given file or directory to a new place.

        :return: Returns the path of the copied file/directory
        :rtype: Path
        """

        logging.debug('Copying "%s" to "%s"', str(self.param), str(self.destination))

        if self.param.is_dir():
            shutil.copytree(self.param, self.destination)
        else:
            shutil.copy2(self.param, self.destination)

        return self.destination



import logging
import os
from pathlib import Path

import click
from github3 import login

from hammurabi import Pillar, __version__
from hammurabi.config import config


@click.group()
@click.pass_context
@click.option(
    "-c",
    "--config",
    "cfg",
    type=click.STRING,
    default="pyproject.toml",
    show_default=True,
    help="Set the configuration file.",
)
@click.option(
    "--repository",
    type=click.STRING,
    default=None,
    help="Set the remote repository. Required format: owner/repository",
)
@click.option(
    "--github-token", type=click.STRING, default=None, help="Set github access token"
)
@click.option(
    "--log-level",
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR"]),
    default=None,
    help="Set logging level.",
)
def cli(
    ctx: click.Context, cfg: str, repository: str, github_token: str, log_level: str
):
    """
    Hammurabi is an extensible CLI tool responsible for enforcing user-defined rules on a git
    repository.

    Find more information at: https://hammurabi.readthedocs.io/latest/
    """

    os.environ.setdefault("HAMMURABI_SETTINGS_PATH", str(Path(cfg).expanduser()))

    try:
        # Reload the configuration
        config.load()
    except Exception as exc:
        raise click.ClickException(str(exc))

    if repository:
        config.settings.repository = repository

    if github_token:
        config.github = login(token=github_token)

    if log_level:
        # Override log level
        logging.root.setLevel(log_level)

    ctx.ensure_object(dict)
    ctx.obj["config"] = config


@cli.command()
def version():
    """
    Print Hammurabi version.
    """

    click.echo(__version__)


@cli.command()
@click.option(
    "-a",
    "--rule-can-abort",
    is_flag=True,
    default=False,
    type=click.BOOL,
    help="Abort the law when a rule raise an exception.",
)
@click.option(
    "--dry-run",
    is_flag=True,
    default=False,
    type=click.BOOL,
    help="Execute in dry run mode.",
)
@click.pass_context
def enforce(ctx: click.Context, rule_can_abort: bool, dry_run: bool):
    """
    Execute all registered Law.
    """

    if dry_run:
        ctx.obj["config"].settings.dry_run = dry_run

    if rule_can_abort:
        ctx.obj["config"].settings.rule_can_abort = rule_can_abort

    pillar: Pillar = ctx.obj["config"].settings.pillar
    pillar.enforce()

    # Generate report for the execution
    pillar.reporter.report()


@cli.group()
def get():
    """
    Show a specific resource or group of resources.
    """


@get.command(name="order")
@click.pass_context
def get_order(ctx: click.Context):
    """
    Show the Pillar's execution order.
    """

    pillar: Pillar = ctx.obj["config"].settings.pillar

    for law in pillar.laws:
        click.echo(f"- {law.name}")

        for rule in law.get_execution_order():
            click.echo(f"  --> {rule.name}")


@get.command(name="laws")
@click.pass_context
def get_laws(ctx: click.Context):
    """
    Show the registered Laws and rules on the Pillar.
    """

    pillar: Pillar = ctx.obj["config"].settings.pillar

    for law in pillar.laws:
        click.echo(f"- {law.name}")


@get.command(name="law", help="Specific law to get")
@click.argument("law", type=click.STRING)
@click.pass_context
def get_law(ctx: click.Context, law: str):
    """
    Show specific registered Law.
    """

    pillar: Pillar = ctx.obj["config"].settings.pillar

    try:
        registered_law = pillar.get_law(law)
    except StopIteration:
        raise click.ClickException(click.style(f'No such law "{law}"', fg="red"))

    click.echo(f"{registered_law.documentation}")


@get.command(name="rules")
@click.pass_context
def get_rules(ctx: click.Context):
    """
    TODO
    """

    pillar: Pillar = ctx.obj["config"].settings.pillar

    for rule in pillar.rules:
        click.echo(f"- {rule.name}")


@get.command(name="rule")
@click.argument("rule", type=click.STRING)
@click.pass_context
def get_rule(ctx: click.Context, rule: str):
    """
    Specific rule to get.
    """

    pillar: Pillar = ctx.obj["config"].settings.pillar

    try:
        registered_rule = pillar.get_rule(rule)
    except StopIteration:
        raise click.ClickException(click.style(f'No such rule "{rule}"', fg="red"))

    click.echo(f"{registered_rule.documentation}")


@cli.group()
def describe():
    """
    Show details of a specific resource or group of resources.
    """


@describe.command(name="law", help="Specific law to describe")
@click.argument("law", type=click.STRING)
@click.pass_context
def describe_law(ctx: click.Context, law: str):
    """
    Show details of the given Law.
    """

    pillar: Pillar = ctx.obj["config"].settings.pillar
    registered_law = pillar.get_law(law)

    if not registered_law:
        raise click.ClickException(click.style(f'No such law "{law}"', fg="red"))

    click.echo(f"{registered_law.documentation}")
    click.echo("Rules:")
    for rule in registered_law.get_execution_order():
        click.echo(f"  --> {rule.name}")


@describe.command(name="rule")
@click.argument("rule", type=click.STRING)
@click.pass_context
def describe_rule(ctx: click.Context, rule: str):
    """
    Specific rule to describe.
    """

    pillar: Pillar = ctx.obj["config"].settings.pillar
    registered_rule = pillar.get_rule(rule)

    if not registered_rule:
        raise click.ClickException(click.style(f'No such rule "{rule}"', fg="red"))

    click.echo(f"{registered_rule.documentation}")

    click.echo("Chain:")
    for chain in registered_rule.get_execution_order():
        click.echo(f"  --> {chain.name}")



"""
This module contains the definition of Rule which describes what to do with
the received parameter and does the necessary changes.

The Rule is an abstract class which describes all the required methods and
parameters, but it can be extended and customized easily by inheriting from
it. A good example for this kind of customization is :class:`hammurabi.rules.text.LineExists`
which adds more parameters to :class:`hammurabi.rules.files.SingleFileRule` which
inherits from :class:`hammurabi.rules.base.Rule`.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
import logging
from typing import Any, Iterable, List, Optional, Union

from hammurabi.config import config
from hammurabi.exceptions import PreconditionFailedError
from hammurabi.preconditions.base import Precondition
from hammurabi.rules.abstract import AbstractRule


class Rule(AbstractRule, ABC):
    """
    Abstract class which describes the bare minimum and helper functions for Rules.
    A rule defines what and how should be executed. Since a rule can have piped and
    children rules, the "parent" rule is responsible for those executions. This kind
    of abstraction allows to run both piped and children rules sequentially in a given
    order.

    Example usage:

    .. code-block:: python

        >>> from typing import Optional
        >>> from pathlib import Path
        >>> from hammurabi import Rule
        >>> from hammurabi.mixins import GitMixin
        >>>
        >>> class SingleFileRule(Rule, GitMixin):
        >>>     def __init__(self, name: str, path: Optional[Path] = None, **kwargs) -> None:
        >>>         super().__init__(name, path, **kwargs)
        >>>
        >>>     def post_task_hook(self):
        >>>         self.git_add(self.param)
        >>>
        >>>     @abstractmethod
        >>>     def task(self) -> Path:
        >>>         pass

    :param name: Name of the rule which will be used for printing
    :type name: str

    :param param: Input parameter of the rule will be used as ``self.param``
    :type param: Any

    :param preconditions: "Boolean Rules" which returns a truthy or falsy value
    :type preconditions: Iterable["Rule"]

    :param pipe: Pipe will be called when the rule is executed successfully
    :type pipe: Optional["Rule"]

    :param children: Children will be executed after the piped rule if there is any
    :type children: Iterable["Rule"]

    .. warning::

        Preconditions can be used in several ways. The most common way is to run
        "Boolean Rules" which takes a parameter and returns a truthy or falsy value.
        In case of a falsy return, the precondition will fail and the rule will not be executed.

        If any modification is done by any of the rules which are used as a
        precondition, those changes will be committed.
    """

    def __init__(
        self,
        name: str,
        param: Any,
        preconditions: Iterable[Precondition] = (),
        pipe: Optional["Rule"] = None,
        children: Iterable["Rule"] = (),
    ) -> None:
        self.pipe = pipe
        self.children = children
        self.preconditions = preconditions

        super().__init__(name, param)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f'name="{self.name}", '
            f'param="{self.param}", '
            f"preconditions={self.preconditions}, "
            f"pipe={self.pipe}, "
            f"children={self.children})"
        )

    def __str__(self) -> str:
        return f"{self.name} rule"

    @property
    def can_proceed(self) -> bool:
        """
        Evaluate if a rule can continue its execution. In case the execution
        is called with ``dry_run`` config option set to true, this method will
        always return ``False`` to make sure not performing any changes. If
        preconditions are set, those will be evaluated by this method.

        :return: Return with the result of evaluation
        :rtype: bool

        .. warning::

            :func:`hammurabi.rules.base.Rule.can_proceed` checks the result of
            ``self.preconditions``, which means the preconditions are executed.
            Make sure that you are not doing any modifications within rules used
            as preconditions, otherwise take extra attention for those rules.
        """

        logging.debug('Checking if "%s" can proceed with execution', self.name)
        proceed: bool = True

        if self.preconditions:
            proceed = all([condition.execute() for condition in self.preconditions])

        return not config.settings.dry_run and proceed

    def __preflight_check(self) -> None:
        """
        Run a preflight check. If the execution can not proceed, determined by
        the can_proceed property, an exception will be raised. We are differentiating
        the exceptions caused by dry run and precondition failure.

        :raises: ``PreconditionFailedError`` if precondition failed otherwise an
            ``AssertionError`` raised
        """

        if not self.can_proceed:
            if config.settings.dry_run:
                raise AssertionError(f'"{self.name}" cannot proceed because of dry run')

            raise PreconditionFailedError(f'"{self.name}" cannot proceed')

    def get_rule_chain(self, rule: "Rule") -> List[Union["Rule", Precondition]]:
        """
        Get the execution chain of the given rule. The execution
        order is the following:

        * task (current rule's :func:`hammurabi.rules.base.Rule.task`)
        * Piped rule
        * Children rules (in the order provided by the iterator used)

        :param rule: The rule which execution chain should be returned
        :type rule: :class:`hammurabi.rules.base.Rule`

        :return: Returns the list of rules in the order above
        :rtype: List[Rule]
        """

        rules: List[Union[Rule, Precondition]] = list(rule.preconditions)
        rules.append(rule)

        if rule.pipe:
            rules.extend(self.get_rule_chain(rule.pipe))

        for child in rule.children:
            rules.extend(self.get_rule_chain(child))

        return rules

    def get_execution_order(self) -> List[Union["Rule", Precondition]]:
        """
        Same as :func:`hammurabi.rules.base.Rule.get_rule_chain` but
        for the root rule.
        """

        order: List[Union[Rule, Precondition]] = list(self.preconditions)
        order.append(self)

        if self.pipe:
            order.extend(self.get_rule_chain(self.pipe))

        for child in self.children:
            order.extend(self.get_rule_chain(child))

        return order

    @abstractmethod
    def task(self) -> Any:
        """
        See the documentation of :func:`hammurabi.rules.base.AbstractRule.task`
        """

    def execute(self, param: Optional[Any] = None):
        """
        Execute the rule's task, its piped and children rules as well.

        The execution order of task, piped rule and children rules
        described in but not by :func:`hammurabi.rules.base.Rule.get_rule_chain`.

        :param param: Input parameter of the rule given by the user
        :type param: Optional[Any]

        :raise: ``AssertionError``
        :return: None

        .. note::

            The input parameter can be optional because of the piped and
            children rules which are receiving the output of its parent. In
            this case the user is not able to set the param manually, since it
            is calculated.

        .. warning::

            If ``self.can_proceed`` returns ``False`` the whole execution
            will be stopped immediately and ``AssertionError`` will be
            raised.
        """

        # In case of piped execution, the Rule will be called without
        # any additional argument, therefore we must set it manually
        # to be able to work with hooks.
        self.param = param or self.param

        self.__preflight_check()

        logging.debug('Running pre task hook for "%s"', self.name)
        self.pre_task_hook()

        logging.info('Running task for "%s"', self.name)
        result = self.task()

        logging.debug('Running post task hook for" %s"', self.name)
        self.post_task_hook()

        logging.info('Rule "%s" finished successfully', self.name)

        if self.pipe:
            logging.debug('Executing pipe "%s" of "%s"', self.pipe.name, self.name)
            self.pipe.execute(result)

        if self.children:
            logging.debug('Executing children rules of "%s"', self.name)
            for child in self.children:
                logging.debug('Executing child "%s" of "%s"', child.name, self.name)
                child.execute(result)



"""
This module contains the definition of the AbstractRule which describes what is shared between
Rules and Preconditions.
"""

from abc import ABC, abstractmethod
from typing import Any, Optional

from hammurabi.helpers import full_strip


class AbstractRule(ABC):
    """
    Abstract class which describes the common behaviour for any kind of rule even
    it is a :class:`hammurabi.rules.base.Rule` or :class:`hammurabi.rules.base.Precondition`

    :param name: Name of the rule which will be used for printing
    :type name: str

    :param param: Input parameter of the rule will be used as ``self.param``
    :type param: Any
    """

    __slots__ = ("param", "name", "made_changes")

    def __init__(self, name: str, param: Any) -> None:
        self.param = param
        self.name = name.strip()

        # Set by GitMixin or other mixins to indicate that the rule did changes.
        # Rules can set this flag directly too. Only those rules will be indicated on
        # Git commit which are made changes.
        self.made_changes = False

    @staticmethod
    def validate(val: Any, cast_to: Optional[Any] = None, required=False) -> Any:
        """

        Validate and/or cast the given value to another type. In case the
        existence of the value is required or casting failed an exception
        will be raised corresponding to the failure.

        :param val: Value to validate
        :type val: Any

        :param cast_to: Type in which the value should be returned
        :type cast_to: Any

        :param required: Check that the value is not falsy
        :type required: bool

        :raise: ``ValueError`` if the given value is required but falsy
        :return: Returns the value in its original or casted type
        :rtype: Any

        Example usage:

        .. code-block:: python

            >>> from typing import Optional
            >>> from pathlib import Path
            >>> from hammurabi import Rule
            >>>
            >>> class MyAwesomeRule(Rule):
            >>>     def __init__(self, name: str, param: Optional[Path] = None):
            >>>         self.param = self.validate(param, required=True)
            >>>
            >>>     # Other method definitions ...
            >>>
        """

        if required and not val:
            raise ValueError("The given value is empty")

        if not cast_to:
            return val

        return cast_to(val)

    @property
    def description(self) -> str:
        """
        Return the description of the :func:`hammurabi.rules.base.Rule.task`
        based on its docstring.

        :return: Stripped description of :func:`hammurabi.rules.base.Rule.task`
        :rtype: str

        .. note::

            As of this property returns the docstring of :func:`hammurabi.rules.base.Rule.task`
            method, it worth to take care of its description when initialized.
        """

        return full_strip(getattr(self.task, "__doc__", ""))

    @property
    def documentation(self) -> str:
        """
        Return the documentation of the rule based on its name, docstring and
        the description of its task.

        :return: Concatenation of the rule's name, docstring, and task description
        :rtype: str

        .. note::

            As of this method returns the name and docstring of the rule
            it worth to take care of its name and description when initialized.
        """

        doc = full_strip(getattr(self, "__doc__", ""))
        return f"{self.name}\n{doc}\n{self.description}"

    def pre_task_hook(self) -> None:
        """
        Run code before performing the :func:`hammurabi.rules.base.Rule.task`.
        To access the parameter passed to the rule, always use ``self.param``
        for :func:`hammurabi.rules.base.Rule.pre_task_hook`.

        .. warning::

            This method is not called in dry run mode.
        """

    def post_task_hook(self):
        """
        Run code after the :func:`hammurabi.rules.base.Rule.task` has been
        performed. To access the parameter passed to the rule, always use
        ``self.param`` for :func:`hammurabi.rules.base.Rule.post_task_hook`.

        .. note::

            This method can be used for execution of git commands
            like git add, or double checking a modification made.

        .. warning::

            This method is not called in dry run mode.
        """

    @abstractmethod
    def task(self) -> Any:
        """
        Abstract method representing how a :func:`hammurabi.rules.base.Rule.task`
        or :func:`hammurabi.preconditions.base.Precondition.task` must be parameterized.
        Any difference in the parameters will result in pylint/mypy errors.

        To be able to use the power of ``pipe`` and ``children``, return
        something which can be generally used for other rules as in input.

        :return: Returns an output which can be used as an input for other rules
        :rtype: Any (usually same as `self.param`'s type)

        .. note::

            Although it is a good practice to return the same type for the output
            that the input has, but this is not the case for "Boolean Rules".
            "Boolean Rules" should return True (or truthy) or False (or falsy) values.
        """



"""
This module contains the definition of Preconditions which are related
to attributes of a file or directory.
"""

from pathlib import Path

from hammurabi.preconditions.base import Precondition


class IsOwnedBy(Precondition):
    """
    Check if the given file or directory has the required ownership.

    To check only the user use ``owner="username"``. To check only the
    group use ``owner=":group_name"`` (please note the colon ``:``).
    It is also possible to check both username and group at the same time
    by using ``owner="username:group_name"``.

    Example usage:

    .. code-block:: python

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, Renamed, IsOwnedBy
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         Renamed(
        >>>             name="Rename pyproject.toml if owned by gabor",
        >>>             path=Path("./pyproject.toml"),
        >>>             new_name="gabor-pyproject.toml"
        >>>             preconditions=[
        >>>                 IsOwnedBy(path=Path("./pyproject.toml"), owner="gabor")
        >>>             ]
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)

    :param path: Input file's path
    :type path: Path

    :param owner: Owner user and/or group of the file/directory separated by colon
    :type owner: str
    """

    def __init__(self, path: Path, owner: str, **kwargs) -> None:
        self.user, self.group = map(lambda x: x.strip(), owner.partition(":")[::2])
        super().__init__(param=path, **kwargs)

    def task(self) -> bool:
        """
        Check if the ownership meets the requirements.

        :return: Returns True if the owner matches
        :rtype: bool
        """

        self.param: Path

        is_owned = False
        is_owned_by_user = self.user == self.param.owner()
        is_owned_by_group = self.group == self.param.group()

        if self.user and self.group:
            is_owned = is_owned_by_user and is_owned_by_group
        elif self.user:
            is_owned = is_owned_by_user
        elif self.group:
            is_owned = is_owned_by_group

        return is_owned


class IsNotOwnedBy(IsOwnedBy):
    """
    Opposite of :class:`hammurabi.preconditions.attributes.IsOwnedBy`.
    """

    def task(self) -> bool:
        """
        Check if the ownership does not meet the requirements.

        :return: Returns True if the owner matches
        :rtype: bool
        """

        return not super().task()


class HasMode(Precondition):
    """
    Check if the given file or directory has the required permissions/mode.

    You can read more about the available modes at https://docs.python.org/3/library/stat.html.

    Example usage:

    .. code-block:: python

        >>> import stat
        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, Renamed, HasMode
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         Renamed(
        >>>             name="Rename pyproject.toml if owned by gabor",
        >>>             path=Path("./pyproject.toml"),
        >>>             new_name="gabor-pyproject.toml"
        >>>             preconditions=[
        >>>                 HasMode(path=Path("scripts/run_unittests.sh"), mode=stat.S_IXOTH)
        >>>             ]
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)

    :param path: Input file's path
    :type path: Path

    :param mode: The desired mode to check
    :type mode: str
    """

    def __init__(self, path: Path, mode: int, **kwargs) -> None:
        self.mode = mode
        super().__init__(param=path, **kwargs)

    def task(self) -> bool:
        """
        Check if the given mode is set on the file or directory.

        :return: Returns True if the desired mode is set
        :rtype: bool
        """

        self.param: Path
        return bool(self.param.stat().st_mode & self.mode)


class HasNoMode(HasMode):
    """
    Opposite of :class:`hammurabi.preconditions.attributes.HasMode`.
    """

    def task(self) -> bool:
        """
        Check if the given mode is not set on the file or directory.

        :return: Returns True if the desired mode is not set
        :rtype: bool
        """

        return not super().task()
